/************************************************************************************
 * Copyright (C) 2018 ALTIMA Company MACNICA, Inc. All Rights Reserved.\n
 *   This software is licensed "AS IS". 
 *   Please perform use of this software by a user's own responsibility and expense.
 *   It cannot guarantee in the maker side about the damage which occurred by the ab-
 *   ility not to use or use this software, and all damage that occurred secondarily.
 ***********************************************************************************/

/* Standard Headers */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>


/* User Headers */
#include "magnes_ioctl.h" //from driver directory
#include "cmd.h"
#include "pci.h"
#include "mem_util.h"


COMMANDS_LIST commands[] = {
		{"menu",	   "menu"                                                           ,cmd_menu},
		{"led",	       "led <period(ms)> <times>"                                       ,cmd_led},
		{"memdump",    "memdump <target(0:ddr/1:ocram)> <offset(HEX)> <size(HEX)>"      ,cmd_mem_dump},
		{"memtest",    "memtest <target(0:ddr/1:ocram)> <mode(0:fixed/1:incremental)>"  ,cmd_mem_test},
		{"pcidump",    "pcidump"                                                        ,cmd_pci_config_dump},
		{"cfgread",    "cfgread <target(0:end/1:root)> <offset(HEX)> "                  ,cmd_cfgread},
		//{"cfgwrite",   "cfgwrite <target(0:end/1:root)> <offset(HEX)> <data(HEX)>"      ,cmd_cfgwrite},
		{"barread",    "barread <BAR:0/1/2..> <offset(HEX)> "                           ,cmd_barX_read},
		{"barwrite",   "barwrite <BAR:0/1/2..> <offset(HEX)> <data(HEX)>"               ,cmd_barX_write},
		{"exit",	"exit"                                                              ,cmd_exit},
		{0,0,cmd_dummy}
};

int cmd_menu(char* options)
{
	int i;

	memset(options,0,1); //Prevent Compiler Warning

	for(i = 0; commands[i].name; i++)  {
		printf(" %-10s:\t%s\n",commands[i].name, commands[i].usage);
	}
	printf(" * Note: HEX Value does not need 0x \n");

	return 0;
}

int cmd_execute(void)
{
	int result = 0;
	int i;
	char input_str[128];

	printf("\n");
	printf("Command: ");
	fflush(stdout);

	fgets(input_str, sizeof(input_str), stdin);
	printf("%s\n", input_str);
	fflush(stdout);

	for(i = 0; commands[i].name; i++) {
		/* parse cmd */
		if(strstr(input_str, commands[i].name) != NULL) {
			/* Execute Command */
			result = commands[i].func(input_str + strlen(commands[i].name));
			break;
		}
	}
	if(commands[i].name == NULL){
		printf("command input error: %s", input_str);
	}

	return result;
}

int cmd_dummy(char* options)
{
	memset(options,0,1); //Prevent Compiler Warning

	printf("CMD Error!! \n");
	return 0;
}

int cmd_exit(char* options)
{
	memset(options,0,1); //Prevent Compiler Warning
	return 1;
}

int cmd_led(char* options)
{
	int i;
	int period_ms;
	int times;
	unsigned int *buf;
	unsigned int size   = 1024;

	printf("led \n");

	if( sscanf(options,"%d %d",&period_ms, &times) < 2)
	{
		printf("arg error \n");
		return 0;
	}


	buf = map_bar(BAR2, BAR2_LED_PIO_BASE, size);
	if(buf == NULL){
		printf("Can't MAP BAR");
	}

	for(i=0; i < times; i++){
		buf[0] = 0x0;
		usleep(1000 * period_ms);
		buf[0] = 0xF;
		usleep(1000 * period_ms);
	}

	munmap(buf, size);

	return 0;
}

int cmd_mem_dump(char* options)
{
	unsigned int target;
	unsigned int offset;
	unsigned int map_offset;
	unsigned int size;
	unsigned int *buf;

	printf("mem_dump \n");

	if( sscanf(options,"%x %x %x",&target, &offset, &size) < 3)
	{
		printf("arg error \n");
		return 0;
	}

	switch(target){
	case 0:
		target = BAR0_DDR_MEM_BASE;
		if( (offset + size) > BAR0_DDR_MEM_SIZE){
			printf("Specified Offset Is Over Actual Memory Space \n");
			return 0;
		}
		break;
	case 1:
		target = BAR0_OCRAM_BASE;
		if( (offset + size) > BAR0_OCRAM_SIZE){
			printf("Specified Offset Is Over Actual Memory Space \n");
			return 0;
		}
		break;
	default:
		printf("target is 0:DDR or 1:OCRAM \n");
		return 0;
	}

	map_offset = offset;
	if(offset > 0){
		if(offset % 0x1000){
			//printf("Specified Offset Is Not 4KB Boundary \n");
			map_offset = offset & 0xFFFFF000;
			//printf("Map Offset is Changed to %x \n",map_offset);
		}
	}

	if(size==0){
		size = 0x100;
	}

	buf = map_bar(BAR0, target + map_offset, size);
	if(buf == NULL){
		printf("Can't MAP BAR");
		return 0;
	}

	sample_memdmp_word(target + offset, buf + ((offset & 0xFFF) >> 2 ), size);

	munmap(buf, size);

	return 0;

}

int cmd_mem_test(char* options)
{

	unsigned int i;
	unsigned int *buf;
	volatile unsigned int read_data;
	unsigned int err = 0;
	unsigned int target;
	unsigned int mode;
	unsigned int size;

	printf("mem_test \n");

	if( sscanf(options,"%x %x",&target, &mode) < 2)
	{
		printf("arg error \n");
		return 0;
	}

	switch(target){
	case 0:
		target = BAR0_DDR_MEM_BASE;
		size   = BAR0_DDR_MEM_SIZE;
		break;
	case 1:
		target = BAR0_OCRAM_BASE;
		size   = BAR0_OCRAM_SIZE;
		break;
	default:
		printf("target is 0:DDR or 1:OCRAM \n");
		return 0;
	}

	buf = map_bar(BAR0, target, size);
	if(buf == NULL){
		printf("Can't MAP BAR");
		return 0;
	}

	switch(mode){
	case 0: //Fixed Data
		printf("Memory Test with Fixed Data Pattern(0xAAAA5555) \n");
		for(i=0; i < size/4UL; i++){
			buf[i] = 0xAAAA5555;
			read_data = buf[i];
			if(read_data != 0xAAAA5555){
				printf("Verify Error at 0x%x: data = 0x%08x \n",target + (i*4), read_data);
				err = 1;
			}
		}
		break;
	case 1: //Incremental Data
		printf("Memory Test with Incremental Data Pattern \n");
		sample_memset_incrementdata(buf, 0, size);
		for(i=0; i < size/4UL; i++){
			read_data = buf[i];
			if(read_data != i){
				printf("Verify Error at 0x%x: data = 0x%08x \n",target + (i*4), read_data);
				err = 1;
			}
		}
		break;
	default:
		printf("Specified mode is not supported \n");
		return 0;
	}

	if(err==0){
		printf("Test Pass!! \n");
	} else {
		printf("Test Fail!! \n");
	}

	printf("Print Top of Target Memory \n");
	sample_memdmp_word(target, buf, 0x100);

	munmap(buf, size);

	return 0;
}

int cmd_cfgread( char *options )
{
	unsigned int target;
	unsigned int offset;
	unsigned int data;

	printf("cfgread \n");

	if( sscanf(options,"%x %x",&target, &offset) < 2)
	{
		printf("arg error \n");
		return 0;
	}

	switch(target)
	{
	case 0:
		printf("cfgread: End \n");
		break;
	case 1:
		printf("cfgread: Root \n");
		break;
	default:
		printf("target is 0:end or 1:root \n");
		return 0;
	}

	config_read(target, offset, &data);
	printf("read_data: %08x \n", data);

	return 0;
}

int cmd_cfgwrite( char *options )
{
	unsigned int offset;
	unsigned int target;
	unsigned int data;

	printf("cfgwrite \n");

	if( sscanf(options,"%x %x %x",&target, &offset, &data) < 3)
	{
		printf("arg error \n");
		return 0;
	}

	switch(target)
	{
	case 0:
		printf("cfgwrite: End \n");
		break;
	case 1:
		printf("cfgwrite: Root \n");
		break;
	default:
		printf("target is 0:end or 1:root \n");
		return 0;
		break;
	}

	config_write(target, offset, data);
	printf("write_data: %08x \n", data);

	return 0;
}



int cmd_pci_config_dump(char* options)
{
	memset(options,0,1); //Prevent Compiler Warning

	//printf("PCI Root Config Space Dump\n");
	//root_pci_config_space_dump();

	printf("PCI End Config Space Dump\n");
	end_pci_config_space_dump();

	return 0;
}

int cmd_barX_read(char* options)
{
	unsigned int bar_num;
	unsigned int offset;
	unsigned int data;

	printf("barX_read \n");

	if( sscanf(options,"%d %x",&bar_num, &offset) < 2)	{
		printf("arg error \n");
		return 0;
	}

	//Select BAR
	switch(bar_num)
	{
	case BAR0:
	case BAR1:
	case BAR2:
	case BAR3:
	case BAR4:
	case BAR5:
		break;
	default:
		printf(" Specified BAR is not correct \n");
		return 0;
	}

	barX_read(bar_num, offset, &data);
	printf("read_data: %08x \n", data);

	return 0;
}

int cmd_barX_write(char* options)
{
	unsigned int bar_num;
	unsigned int offset;
	unsigned int data;

	printf("barX_write \n");

	if( sscanf(options,"%d %x %x",&bar_num, &offset, &data) < 3) {
		printf("arg error \n");
		return 0;
	}
	//Select BAR
	switch(bar_num)
	{
	case BAR0:
	case BAR1:
	case BAR2:
	case BAR3:
	case BAR4:
	case BAR5:
		break;
	default:
		printf(" Specified BAR is not correct \n");
		return 0;
	}

	barX_write(bar_num, offset, data);
	printf("write_data: %08x \n", data);

	return 0;
}

