/* 
 * Copyright (c) 2017, Intel Corporation.
 * Intel, the Intel logo, Intel, MegaCore, NIOS II, Quartus and TalkBack 
 * words and logos are trademarks of Intel Corporation or its subsidiaries 
 * in the U.S. and/or other countries. Other marks and brands may be 
 * claimed as the property of others.   See Trademarks on intel.com for 
 * full list of Intel trademarks or the Trademarks & Brands Names Database 
 * (if Intel) or See www.Intel.com/legal (if Altera).
 * All rights reserved
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * BSD 3-Clause license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 *      - Neither Intel nor the names of its contributors may be 
 *        used to endorse or promote products derived from this 
 *        software without specific prior written permission.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/************************************************************************************
 * Copyright (C) 2018 ALTIMA Company MACNICA, Inc. All Rights Reserved.\n
 *   This software is licensed "AS IS". 
 *   Please perform use of this software by a user's own responsibility and expense.
 *   It cannot guarantee in the maker side about the damage which occurred by the ab-
 *   ility not to use or use this software, and all damage that occurred secondarily.
 ***********************************************************************************/

#include  <linux/version.h>

#include "magnes_pci.h"
#include "magnes_ioctl.h"

#ifndef VM_RESERVED
#define VM_RESERVED (VM_DONTEXPAND | VM_DONTDUMP)
#endif


volatile int bar_num_select = 0;

int magnes_open(struct inode *inode, struct file *file)
{
	struct magnes_pci_dev *magnes_pci = 0;
	magnes_pci = container_of(inode->i_cdev, struct magnes_pci_dev, cdev);
	file->private_data = magnes_pci;

	printk("magnes_open \n");

	return 0;
}

int magnes_close(struct inode *inode, struct file *file)
{
	printk("magnes_close \n");
	return 0;
}

ssize_t magnes_read(struct file *filp, char __user *buf, size_t count, loff_t *f_pos)
{
	int retval;
	struct magnes_pci_dev *magnes_pci = filp->private_data;
	struct magnes_rw_cmd cmd;
	unsigned int *ptr;


	printk("magnes_read \n");
	retval = copy_from_user (&cmd, (int __user *)buf, sizeof(cmd));

	//cmd.data = *(unsigned int *)((void *)(magnes_pci->bar[cmd.bar_num] + cmd.offset));
	//ptr = (unsigned int *)(magnes_pci->bar[cmd.bar_num] + cmd.offset);
	//ptr = (unsigned int *)(magnes_pci->bar[cmd.bar_num]);
	ptr = (unsigned int *)(magnes_pci->bar[2]);
	//printk("BAR[%d]: at 0x%p \n", cmd.bar_num, ptr);
	//cmd.data = ptr[0x4000 >> 2];
	//cmd.data = readl(ptr);
	retval = copy_to_user ((int __user *)buf, &cmd, sizeof(cmd));
#if 0
	switch(cmd.bar_num){
	case BAR0:
		magnes_pci->bar[BAR0 + cmd.offset] = cmd.data;
		break;
	case BAR2:
		magnes_pci->bar[BAR2 + cmd.offset] = cmd.data;
		break;
	default:
		break;

	}
#endif

	return 1;
}


ssize_t magnes_write(struct file *filp, const char __user *buf, size_t count, loff_t *f_pos)
{
	int retval;
	struct magnes_pci_dev *magnes_pci = filp->private_data;
	struct magnes_rw_cmd cmd;
	unsigned int *ptr;

	printk("magnes_write \n");

	retval = copy_from_user (&cmd, (int __user *)buf, sizeof(cmd));

	//ptr = (void *)(magnes_pci->bar[cmd.bar_num] + cmd.offset);
	ptr = (void *)(magnes_pci->bar[cmd.bar_num]);
	cmd.data = (unsigned int)(ptr[cmd.offset]);

#if 0
	switch(cmd.bar_num){
	case BAR0:
		//cmd.data = magnes_pci->bar[BAR0 + cmd.offset];
		cmd.data = *(magnes_pci->bar[cmd.bar_num] + cmd.offset);
		break;
	case BAR2:
		//cmd.data = magnes_pci->bar[BAR2 + cmd.offset];
		break;
	default:
		break;

	}
#endif
	retval = copy_to_user ((int __user *)buf, &cmd, sizeof(cmd));

	return 1;
}

long magnes_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	int retval;
	struct magnes_ioctl_cmd read_cmd;
	struct magnes_ioctl_cmd write_cmd;

	struct magnes_pci_dev *magnes_pci = file->private_data;
	struct pci_dev *pci_dev;

	unsigned int *target_bar;

	printk("magnes_ioctl \n");
	printk("cmd = %d \n", _IOC_NR(cmd));

	switch(cmd) {
	case MAGNES_CMD_CONFIG_READ:
		printk("MAGNES_CMD_CONFIG_READ \n");
		retval = copy_from_user (&read_cmd, (int __user *)arg, sizeof(read_cmd));

		if(read_cmd.cfg_num == 0){
			//End Config
			pci_dev = magnes_pci->pci_dev;
			retval = pci_read_config_dword(pci_dev, read_cmd.cfg_offset, &read_cmd.buf);
		}else {
			//Root Config
			pci_dev = magnes_pci->upstream;
			retval = pci_read_config_dword(pci_dev, read_cmd.cfg_offset, &read_cmd.buf);
		}
		if(retval != 0){
			printk("Error: pci_read_config_dword: %d \n",retval);
		}

		retval = copy_to_user ((int __user *)arg, &read_cmd, sizeof(read_cmd));

		printk("read_cmd.buf = %x \n", read_cmd.buf);

		break;

	case MAGNES_CMD_CONFIG_WRITE:
		printk("MAGNES_CMD_CONFIG_WRITE \n");
		retval = copy_from_user (&write_cmd, (int __user *)arg, sizeof(write_cmd));

		if(write_cmd.cfg_num == 0){
			//End Config
			pci_dev = magnes_pci->pci_dev;
			retval = pci_write_config_dword(pci_dev, write_cmd.cfg_offset, write_cmd.buf);
		}else {
			//Root Config
			pci_dev = magnes_pci->upstream;
			retval = pci_write_config_dword(pci_dev, write_cmd.cfg_offset, write_cmd.buf);
		}
		if(retval != 0){
			printk("Error: pci_write_config_dword: %d \n",retval);
		}

		retval = copy_to_user ((int __user *)arg, &write_cmd, sizeof(write_cmd));

		printk("write_cmd.buf = %x \n", write_cmd.buf);

		break;

	case MAGNES_CMD_BAR_READ:
		printk("MAGNES_CMD_BAR_READ \n");

		retval = copy_from_user (&read_cmd, (int __user *)arg, sizeof(read_cmd));

		target_bar = magnes_pci->bar[read_cmd.bar_num];

		read_cmd.buf = target_bar[read_cmd.bar_offset >> 2];

		retval = copy_to_user ((int __user *)arg, &read_cmd, sizeof(read_cmd));
		printk("read_cmd.buf = %x \n", read_cmd.buf);

		break;

	case MAGNES_CMD_BAR_WRITE:
		printk("MAGNES_CMD_BAR_WRITE \n");
		retval = copy_from_user (&write_cmd, (int __user *)arg, sizeof(write_cmd));

		target_bar = magnes_pci->bar[write_cmd.bar_num];
		target_bar[write_cmd.bar_offset >> 2] = write_cmd.buf;

		retval = copy_to_user ((int __user *)arg, &write_cmd, sizeof(write_cmd));
		printk("write_cmd.buf = %x \n", write_cmd.buf);

		break;

	case MAGNES_CMD_BAR_MAP:
		printk("MAGNES_CMD_BAR_MAP \n");
		retval = copy_from_user (&write_cmd, (int __user *)arg, sizeof(write_cmd));
		magnes_pci->mmap_target = write_cmd.bar_num;
		retval = copy_to_user ((int __user *)arg, &write_cmd, sizeof(write_cmd));

		break;

	default:
		printk("Pass Here?? \n");
		break;
	}

	return 1;
}


int magnes_mmap(struct file *file, struct vm_area_struct *p_vm)
{
	int ret;
	unsigned long map_size = p_vm->vm_end - p_vm->vm_start;
	unsigned long offset = p_vm->vm_pgoff << PAGE_SHIFT;
	unsigned long mmap_physaddr;
	unsigned long mappable_size;
	struct magnes_pci_dev *magnes_pci = file->private_data;

#if 1
	switch (magnes_pci->mmap_target) {
	case 0:
	case 1:
	case 2:
	case 3:
	case 4:
	case 5:
		//mmap_physaddr = bar2_addr + offset; //グローバルで持って来ちゃったので構造体に入れ込む
		mmap_physaddr = magnes_pci->bar_phys_addr[magnes_pci->mmap_target] + offset; //グローバルで持って来ちゃったので構造体に入れ込む
		mappable_size = magnes_pci->bar_length[magnes_pci->mmap_target];
		p_vm->vm_flags |= VM_RESERVED | VM_IO;
		break;
	default:
		printk("Invalid mmap target: %d \n", magnes_pci->mmap_target);
		return EINVAL;
	}


#else
	switch (magnes_pci_dev->mmap_target) {

	case ALTDMAC_MMAP_REGS:
		mmap_physaddr = p_altdmac->pciconf.base_address[ALTDMAC_PCI_BAR_REGS] + offset;
		mappable_size = p_altdmac->pciconf.base_address_size[ALTDMAC_PCI_BAR_REGS];
		p_vm->vm_flags |= VM_RESERVED | VM_IO;
		break;

	case ALTDMAC_MMAP_IMEM:
		mmap_physaddr = p_altdmac->pciconf.base_address[ALTDMAC_PCI_BAR_IMEM] + offset;
		mappable_size = p_altdmac->pciconf.base_address_size[ALTDMAC_PCI_BAR_IMEM];
		p_vm->vm_flags |= VM_RESERVED;
		break;

	case ALTDMAC_MMAP_DMARBUF:
		mmap_physaddr = p_altdmac->dmarbuf_physaddr + offset;
		mappable_size = ALTDMAC_DMA_MAX_SIZE;
		p_vm->vm_flags |= VM_RESERVED;
		break;

	case ALTDMAC_MMAP_DMAWBUF:
		mmap_physaddr = p_altdmac->dmawbuf_physaddr + offset;
		mappable_size = ALTDMAC_DMA_MAX_SIZE;
		p_vm->vm_flags |= VM_RESERVED;
		break;

	case ALTDMAC_MMAP_DMADESC_TABLE:
		mmap_physaddr = p_altdmac->dmardesc_table_physaddr + offset;
		mappable_size = sizeof(dmadesc_table_t) * 2; // for read and write
		p_vm->vm_flags |= VM_RESERVED;
		break;

	default:
		return -EINVAL;
	}

#endif
	if (offset > mappable_size)
		return -EINVAL;

	/********************************************************
	 * IO memory or DMA memory area may look like corrupted *
	 * in user space if not marked as non cache area.       *
	 * Actually, without the following code, I confimed     *
	 * that read data from I/O mapped area (DMA registers   *
	 * area) had been corrupt. Because dma_mmap_coherent()  *
	 * function use this pgprot_noncached() function, I put *
	 * this code and the corrupt issue was solved.          *
	 ********************************************************/
#ifdef pgprot_noncached ////////////////////////////////////
	// marked as non chache area
	p_vm->vm_page_prot = pgprot_noncached(p_vm->vm_page_prot);
#endif /////////////////////////////////////pgprot_noncached

	ret = remap_pfn_range(p_vm,
	                      p_vm->vm_start,
	                      mmap_physaddr >> PAGE_SHIFT,
	                      map_size,
	                      p_vm->vm_page_prot);
	if (ret)
		return -EAGAIN;

	return 0;
}
